/**
 * CAROL: Common Architecture for RMI Objectweb Layer
 * Copyright (C) 2009 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: InterceptableContext.java 2165 2009-05-06 07:26:22Z benoitf $
 * --------------------------------------------------------------------------
 */

package org.ow2.carol.jndi.intercept.spi;

import java.util.Hashtable;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.ow2.carol.jndi.intercept.ContextInterceptor;
import org.ow2.carol.jndi.intercept.InterceptorManager;
import org.ow2.carol.jndi.intercept.operation.AbstractInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.AddToEnvironmentInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.CloseInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.GetEnvironmentInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.GetNameInNamespaceInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameBindInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameComposeNameInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameCreateSubcontextInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameDestroySubcontextInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameGetNameParserInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameListBindingsInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameListInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameLookupInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameLookupLinkInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameRebindInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameRenameInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.NameUnbindInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.RemoveFromEnvironmentInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringBindInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringComposeNameInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringCreateSubcontextInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringDestroySubcontextInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringGetNameParserInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringListBindingsInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringListInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringLookupInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringLookupLinkInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringRebindInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringRenameInterceptionContext;
import org.ow2.carol.jndi.intercept.operation.StringUnbindInterceptionContext;

/**
 * The InterceptorContext is ...
 *
 * @author Guillaume Sauthier
 */
public class InterceptableContext implements Context {


    private InterceptorManager manager;
    private Context wrapped;

    public InterceptableContext(final InterceptorManager manager,
                                final Context wrapped) {
        this.manager = manager;
        this.wrapped = wrapped;
    }

    /**
     * Retrieves the named object.
     * If <tt>name</tt> is empty, returns a new instance of this context
     * (which represents the same naming context as this context, but its
     * environment may be modified independently and it may be accessed
     * concurrently).
     *
     * @param name the name of the object to look up
     * @return the object bound to <tt>name</tt>
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #lookup(String)
     * @see #lookupLink(javax.naming.Name)
     */
    public Object lookup(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new NameLookupInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform lookup", e);
        }
    }

    /**
     * Retrieves the named object.
     * See {@link #lookup(javax.naming.Name)} for details.
     *
     * @param name the name of the object to look up
     * @return the object bound to <tt>name</tt>
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public Object lookup(final String name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new StringLookupInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform lookup", e);
        }
    }

    /**
     * Binds a name to an object.
     * All intermediate contexts and the target context (that named by all
     * but terminal atomic component of the name) must already exist.
     *
     * @param name the name to bind; may not be empty
     * @param obj  the object to bind; possibly null
     * @throws javax.naming.NameAlreadyBoundException if name is already bound
     * @throws javax.naming.directory.InvalidAttributesException if object did not supply all mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #bind(String, Object)
     * @see #rebind(javax.naming.Name , Object)
     * @see javax.naming.directory.DirContext#bind(javax.naming.Name , Object,
     *      javax.naming.directory.Attributes)
     */
    public void bind(final Name name, final Object obj) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameBindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name, obj});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform bind", e);
        }
    }

    /**
     * Binds a name to an object.
     * See {@link #bind(javax.naming.Name , Object)} for details.
     *
     * @param name the name to bind; may not be empty
     * @param obj  the object to bind; possibly null
     * @throws javax.naming.NameAlreadyBoundException if name is already bound
     * @throws javax.naming.directory.InvalidAttributesException if object did not supply all mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void bind(final String name, final Object obj) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringBindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name, obj});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform bind", e);
        }
    }

    /**
     * Binds a name to an object, overwriting any existing binding.
     * All intermediate contexts and the target context (that named by all
     * but terminal atomic component of the name) must already exist.
     * <p/>
     * <p> If the object is a <tt>DirContext</tt>, any existing attributes
     * associated with the name are replaced with those of the object.
     * Otherwise, any existing attributes associated with the name remain
     * unchanged.
     *
     * @param name the name to bind; may not be empty
     * @param obj  the object to bind; possibly null
     * @throws javax.naming.directory.InvalidAttributesException if object did not supply all mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #rebind(String, Object)
     * @see #bind(javax.naming.Name , Object)
     * @see javax.naming.directory.DirContext#rebind(javax.naming.Name , Object,
     *      javax.naming.directory.Attributes)
     * @see javax.naming.directory.DirContext
     */
    public void rebind(final Name name, final Object obj) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameRebindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name, obj});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform rebind", e);
        }
    }

    /**
     * Binds a name to an object, overwriting any existing binding.
     * See {@link #rebind(javax.naming.Name , Object)} for details.
     *
     * @param name the name to bind; may not be empty
     * @param obj  the object to bind; possibly null
     * @throws javax.naming.directory.InvalidAttributesException if object did not supply all mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void rebind(final String name, final Object obj) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringRebindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name, obj});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform rebind", e);
        }
    }

    /**
     * Unbinds the named object.
     * Removes the terminal atomic name in <code>name</code>
     * from the target context--that named by all but the terminal
     * atomic part of <code>name</code>.
     * <p/>
     * <p> This method is idempotent.
     * It succeeds even if the terminal atomic name
     * is not bound in the target context, but throws
     * <tt>NameNotFoundException</tt>
     * if any of the intermediate contexts do not exist.
     * <p/>
     * <p> Any attributes associated with the name are removed.
     * Intermediate contexts are not changed.
     *
     * @param name the name to unbind; may not be empty
     * @throws javax.naming.NameNotFoundException if an intermediate context does not exist
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #unbind(String)
     */
    public void unbind(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameUnbindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform unbind", e);
        }
    }

    /**
     * Unbinds the named object.
     * See {@link #unbind(javax.naming.Name)} for details.
     *
     * @param name the name to unbind; may not be empty
     * @throws javax.naming.NameNotFoundException if an intermediate context does not exist
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void unbind(final String name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringUnbindInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform unbind", e);
        }
    }

    /**
     * Binds a new name to the object bound to an old name, and unbinds
     * the old name.  Both names are relative to this context.
     * Any attributes associated with the old name become associated
     * with the new name.
     * Intermediate contexts of the old name are not changed.
     *
     * @param oldName the name of the existing binding; may not be empty
     * @param newName the name of the new binding; may not be empty
     * @throws javax.naming.NameAlreadyBoundException if <tt>newName</tt> is already bound
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #rename(String, String)
     * @see #bind(javax.naming.Name , Object)
     * @see #rebind(javax.naming.Name , Object)
     */
    public void rename(final Name oldName, final Name newName) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameRenameInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {oldName, newName});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform rename", e);
        }
    }

    /**
     * Binds a new name to the object bound to an old name, and unbinds
     * the old name.
     * See {@link #rename(javax.naming.Name , javax.naming.Name)} for details.
     *
     * @param oldName the name of the existing binding; may not be empty
     * @param newName the name of the new binding; may not be empty
     * @throws javax.naming.NameAlreadyBoundException if <tt>newName</tt> is already bound
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void rename(final String oldName, final String newName) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringRenameInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {oldName, newName});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform rename", e);
        }
    }

    /**
     * Enumerates the names bound in the named context, along with the
     * class names of objects bound to them.
     * The contents of any subcontexts are not included.
     * <p/>
     * <p> If a binding is added to or removed from this context,
     * its effect on an enumeration previously returned is undefined.
     *
     * @param name the name of the context to list
     * @return an enumeration of the names and class names of the
     * bindings in this context.  Each element of the
     * enumeration is of type <tt>NameClassPair</tt>.
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #list(String)
     * @see #listBindings(javax.naming.Name)
     * @see javax.naming.NameClassPair
     */
    @SuppressWarnings("unchecked")
	public NamingEnumeration<NameClassPair> list(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameListInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (NamingEnumeration<NameClassPair>) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform list", e);
        }
    }

    /**
     * Enumerates the names bound in the named context, along with the
     * class names of objects bound to them.
     * See {@link #list(javax.naming.Name)} for details.
     *
     * @param name the name of the context to list
     * @return an enumeration of the names and class names of the
     * bindings in this context.  Each element of the
     * enumeration is of type <tt>NameClassPair</tt>.
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    @SuppressWarnings("unchecked")
	public NamingEnumeration<NameClassPair> list(final String name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringListInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (NamingEnumeration<NameClassPair>) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform list", e);
        }
    }

    /**
     * Enumerates the names bound in the named context, along with the
     * objects bound to them.
     * The contents of any subcontexts are not included.
     * <p/>
     * <p> If a binding is added to or removed from this context,
     * its effect on an enumeration previously returned is undefined.
     *
     * @param name the name of the context to list
     * @return an enumeration of the bindings in this context.
     * Each element of the enumeration is of type
     * <tt>Binding</tt>.
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #listBindings(String)
     * @see #list(javax.naming.Name)
     * @see javax.naming.Binding
     */
    @SuppressWarnings("unchecked")
	public NamingEnumeration<Binding> listBindings(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameListBindingsInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (NamingEnumeration<Binding>) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform listBindings", e);
        }
    }

    /**
     * Enumerates the names bound in the named context, along with the
     * objects bound to them.
     * See {@link #listBindings(javax.naming.Name)} for details.
     *
     * @param name the name of the context to list
     * @return an enumeration of the bindings in this context.
     * Each element of the enumeration is of type
     * <tt>Binding</tt>.
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    @SuppressWarnings("unchecked")
	public NamingEnumeration<Binding> listBindings(final String name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringListBindingsInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (NamingEnumeration<Binding>) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform listBindings", e);
        }
    }

    /**
     * Destroys the named context and removes it from the namespace.
     * Any attributes associated with the name are also removed.
     * Intermediate contexts are not destroyed.
     * <p/>
     * <p> This method is idempotent.
     * It succeeds even if the terminal atomic name
     * is not bound in the target context, but throws
     * <tt>NameNotFoundException</tt>
     * if any of the intermediate contexts do not exist.
     * <p/>
     * <p> In a federated naming system, a context from one naming system
     * may be bound to a name in another.  One can subsequently
     * look up and perform operations on the foreign context using a
     * composite name.  However, an attempt destroy the context using
     * this composite name will fail with
     * <tt>NotContextException</tt>, because the foreign context is not
     * a "subcontext" of the context in which it is bound.
     * Instead, use <tt>unbind()</tt> to remove the
     * binding of the foreign context.  Destroying the foreign context
     * requires that the <tt>destroySubcontext()</tt> be performed
     * on a context from the foreign context's "native" naming system.
     *
     * @param name the name of the context to be destroyed; may not be empty
     * @throws javax.naming.NameNotFoundException if an intermediate context does not exist
     * @throws javax.naming.NotContextException if the name is bound but does not name a
     * context, or does not name a context of the appropriate type
     * @throws javax.naming.ContextNotEmptyException if the named context is not empty
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #destroySubcontext(String)
     */
    public void destroySubcontext(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameDestroySubcontextInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform destroySubcontext", e);
        }
    }

    /**
     * Destroys the named context and removes it from the namespace.
     * See {@link #destroySubcontext(javax.naming.Name)} for details.
     *
     * @param name the name of the context to be destroyed; may not be empty
     * @throws javax.naming.NameNotFoundException if an intermediate context does not exist
     * @throws javax.naming.NotContextException if the name is bound but does not name a
     * context, or does not name a context of the appropriate type
     * @throws javax.naming.ContextNotEmptyException if the named context is not empty
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void destroySubcontext(final String name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringDestroySubcontextInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform destroySubcontext", e);
        }
    }

    /**
     * Creates and binds a new context.
     * Creates a new context with the given name and binds it in
     * the target context (that named by all but terminal atomic
     * component of the name).  All intermediate contexts and the
     * target context must already exist.
     *
     * @param name the name of the context to create; may not be empty
     * @return the newly created context
     * @throws javax.naming.NameAlreadyBoundException if name is already bound
     * @throws javax.naming.directory.InvalidAttributesException if creation of the subcontext requires specification of
     * mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #createSubcontext(String)
     * @see javax.naming.directory.DirContext#createSubcontext
     */
    public Context createSubcontext(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new NameCreateSubcontextInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (Context) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform createSubcontext", e);
        }
    }

    /**
     * Creates and binds a new context.
     * See {@link #createSubcontext(javax.naming.Name)} for details.
     *
     * @param name the name of the context to create; may not be empty
     * @return the newly created context
     * @throws javax.naming.NameAlreadyBoundException if name is already bound
     * @throws javax.naming.directory.InvalidAttributesException if creation of the subcontext requires specification of
     * mandatory attributes
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public Context createSubcontext(final String name) throws NamingException {
        try {
            AbstractInterceptionContext intercepted = new StringCreateSubcontextInterceptionContext(wrapped);
            intercepted.setInterceptors(getInterceptors().iterator());
            intercepted.setParameters(new Object[] {name});
            return (Context) intercepted.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform createSubcontext", e);
        }
    }

    /**
     * Retrieves the named object, following links except
     * for the terminal atomic component of the name.
     * If the object bound to <tt>name</tt> is not a link,
     * returns the object itself.
     *
     * @param name the name of the object to look up
     * @return the object bound to <tt>name</tt>, not following the
     * terminal link (if any).
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #lookupLink(String)
     */
    public Object lookupLink(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new NameLookupLinkInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform lookupLink", e);
        }
    }

    /**
     * Retrieves the named object, following links except
     * for the terminal atomic component of the name.
     * See {@link #lookupLink(javax.naming.Name)} for details.
     *
     * @param name the name of the object to look up
     * @return the object bound to <tt>name</tt>, not following the
     * terminal link (if any)
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public Object lookupLink(final String name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new StringLookupLinkInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform lookupLink", e);
        }
    }

    /**
     * Retrieves the parser associated with the named context.
     * In a federation of namespaces, different naming systems will
     * parse names differently.  This method allows an application
     * to get a parser for parsing names into their atomic components
     * using the naming convention of a particular naming system.
     * Within any single naming system, <tt>NameParser</tt> objects
     * returned by this method must be equal (using the <tt>equals()</tt>
     * test).
     *
     * @param name the name of the context from which to get the parser
     * @return a name parser that can parse compound names into their atomic
     * components
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #getNameParser(String)
     * @see javax.naming.CompoundName
     */
    public NameParser getNameParser(final Name name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new NameGetNameParserInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return (NameParser) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform getNameParser", e);
        }
    }

    /**
     * Retrieves the parser associated with the named context.
     * See {@link #getNameParser(javax.naming.Name)} for details.
     *
     * @param name the name of the context from which to get the parser
     * @return a name parser that can parse compound names into their atomic
     * components
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public NameParser getNameParser(final String name) throws NamingException {
        try {
            AbstractInterceptionContext lc = new StringGetNameParserInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name});
            return (NameParser) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform getNameParser", e);
        }
    }

    /**
     * Composes the name of this context with a name relative to
     * this context.
     * Given a name (<code>name</code>) relative to this context, and
     * the name (<code>prefix</code>) of this context relative to one
     * of its ancestors, this method returns the composition of the
     * two names using the syntax appropriate for the naming
     * system(s) involved.  That is, if <code>name</code> names an
     * object relative to this context, the result is the name of the
     * same object, but relative to the ancestor context.  None of the
     * names may be null.
     * <p/>
     * For example, if this context is named "wiz.com" relative
     * to the initial context, then
     * <pre>
     * 	composeName("east", "wiz.com")	</pre>
     * might return <code>"east.wiz.com"</code>.
     * If instead this context is named "org/research", then
     * <pre>
     * 	composeName("user/jane", "org/research")	</pre>
     * might return <code>"org/research/user/jane"</code> while
     * <pre>
     * 	composeName("user/jane", "research")	</pre>
     * returns <code>"research/user/jane"</code>.
     *
     * @param name   a name relative to this context
     * @param prefix the name of this context relative to one of its ancestors
     * @return the composition of <code>prefix</code> and <code>name</code>
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #composeName(String, String)
     */
    public Name composeName(final Name name, final Name prefix) throws NamingException {
        try {
            AbstractInterceptionContext lc = new NameComposeNameInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name, prefix});
            return (Name) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform composeName", e);
        }
    }

    /**
     * Composes the name of this context with a name relative to
     * this context.
     * See {@link #composeName(javax.naming.Name , javax.naming.Name)} for details.
     *
     * @param name   a name relative to this context
     * @param prefix the name of this context relative to one of its ancestors
     * @return the composition of <code>prefix</code> and <code>name</code>
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public String composeName(final String name, final String prefix) throws NamingException {
        try {
            AbstractInterceptionContext lc = new StringComposeNameInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {name, prefix});
            return (String) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform composeName", e);
        }
    }

    /**
     * Adds a new environment property to the environment of this
     * context.  If the property already exists, its value is overwritten.
     * See class description for more details on environment properties.
     *
     * @param propName the name of the environment property to add; may not be null
     * @param propVal  the value of the property to add; may not be null
     * @return the previous value of the property, or null if the property was
     * not in the environment before
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #getEnvironment()
     * @see #removeFromEnvironment(String)
     */
    public Object addToEnvironment(final String propName, final Object propVal) throws NamingException {
        try {
            AbstractInterceptionContext lc = new AddToEnvironmentInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {propName, propVal});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform addToEnvironment", e);
        }
    }

    /**
     * Removes an environment property from the environment of this
     * context.  See class description for more details on environment
     * properties.
     *
     * @param propName the name of the environment property to remove; may not be null
     * @return the previous value of the property, or null if the property was
     * not in the environment
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #getEnvironment()
     * @see #addToEnvironment(String, Object)
     */
    public Object removeFromEnvironment(final String propName) throws NamingException {
        try {
            AbstractInterceptionContext lc = new RemoveFromEnvironmentInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.setParameters(new Object[] {propName});
            return lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform removeFromEnvironment", e);
        }
    }

    /**
     * Retrieves the environment in effect for this context.
     * See class description for more details on environment properties.
     * <p/>
     * <p> The caller should not make any changes to the object returned:
     * their effect on the context is undefined.
     * The environment of this context may be changed using
     * <tt>addToEnvironment()</tt> and <tt>removeFromEnvironment()</tt>.
     *
     * @return the environment of this context; never null
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @see #addToEnvironment(String, Object)
     * @see #removeFromEnvironment(String)
     */
    public Hashtable<?, ?> getEnvironment() throws NamingException {
        try {
            AbstractInterceptionContext lc = new GetEnvironmentInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            return (Hashtable<?, ?>) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform getEnvironment", e);
        }
    }

    /**
     * Closes this context.
     * This method releases this context's resources immediately, instead of
     * waiting for them to be released automatically by the garbage collector.
     * <p/>
     * <p> This method is idempotent:  invoking it on a context that has
     * already been closed has no effect.  Invoking any other method
     * on a closed context is not allowed, and results in undefined behaviour.
     *
     * @throws javax.naming.NamingException if a naming exception is encountered
     */
    public void close() throws NamingException {
        try {
            AbstractInterceptionContext lc = new CloseInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform close", e);
        }
    }

    /**
     * Retrieves the full name of this context within its own namespace.
     * <p/>
     * <p> Many naming services have a notion of a "full name" for objects
     * in their respective namespaces.  For example, an LDAP entry has
     * a distinguished name, and a DNS record has a fully qualified name.
     * This method allows the client application to retrieve this name.
     * The string returned by this method is not a JNDI composite name
     * and should not be passed directly to context methods.
     * In naming systems for which the notion of full name does not
     * make sense, <tt>OperationNotSupportedException</tt> is thrown.
     *
     * @return this context's name in its own namespace; never null
     * @throws javax.naming.OperationNotSupportedException if the naming system does
     * not have the notion of a full name
     * @throws javax.naming.NamingException if a naming exception is encountered
     * @since 1.3
     */
    public String getNameInNamespace() throws NamingException {
        try {
            AbstractInterceptionContext lc = new GetNameInNamespaceInterceptionContext(wrapped);
            lc.setInterceptors(getInterceptors().iterator());
            return (String) lc.proceed();
        } catch (Exception e) {
            throw namingException("Cannot perform getNameInNamespace", e);
        }
    }


    private Iterable<ContextInterceptor> getInterceptors() {
        return manager.getInterceptors();
    }

    private static NamingException namingException(final String message, final Exception cause) {
        NamingException ne = new NamingException(message);
        return (NamingException) ne.initCause(cause);
    }

}
