/**
 * Copyright (C) 2008-2009 - Bull S.A.S.
 *
 * CAROL: Common Architecture for RMI ObjectWeb Layer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: CMIConfigurationHelper.java 2206 2010-02-24 16:21:55Z benoitf $
 * --------------------------------------------------------------------------
 */
package org.ow2.carol.util.configuration;

import java.net.URL;
import java.util.ArrayList;
import java.util.Hashtable;
import java.util.List;
import java.util.Properties;
import java.util.Set;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.ow2.cmi.admin.MBeanUtils;
import org.ow2.cmi.config.CMIConfig;
import org.ow2.cmi.config.CMIConfigException;
import org.ow2.cmi.config.CMIConfigurator;
import org.ow2.cmi.config.CMIProperty;
import org.ow2.cmi.config.JNDIConfig;
import org.ow2.cmi.config.ProtocolConfig;
import org.ow2.cmi.config.ProviderUrls;
import org.ow2.cmi.controller.common.IConfig;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactory;
import org.ow2.cmi.controller.factory.ClusterViewManagerFactoryConfig;
import org.ow2.cmi.jndi.context.CMIInitialContextFactory;

/**
 * Helper to enable CMI in Carol.
 * @author Loris Bouzonnet
 */
public final class CMIConfigurationHelper {

    /**
     * Utility class: no public constructor.
     */
    private CMIConfigurationHelper() {
    }

    /**
     * Logger.
     */
    private static Log logger = LogFactory.getLog(CMIConfigurationHelper.class);


    /**
     * @param sc
     * @param domainName
     * @param serverName
     * @param agentId
     * @param cmiConfigUrl
     * @throws ConfigurationException
     */
    protected static boolean enableCMI(
            final ServerConfiguration sc, final String domainName,
            final String serverName, final String agentId, final URL cmiConfigUrl) throws ConfigurationException {
        CMIConfig<?> cmiConfig;
        if(cmiConfigUrl == null) {
            Properties cmiEnv;
            ProtocolConfiguration[] protocolConfigurations = ConfigurationRepository.getConfigurations();
            if(protocolConfigurations.length != 1) {
                logger.debug("Many protocols: CMI client only supports one protocol so you need to specify one of them.");
                return false;
            }
            cmiEnv = getDefaultCMIEnv(protocolConfigurations[0].getProtocol().getName());
            try {
                cmiConfig = CMIConfigurator.extractCMIConfig(cmiEnv, false);
            } catch (CMIConfigException e) {
                logger.error("Unable to extract a configuration for CMI", e);
                throw new ConfigurationException(
                        "Unable to extract a configuration for CMI", e);
            }
        } else {
            try {
                cmiConfig = CMIConfigurator.extractCMIConfig(cmiConfigUrl);
            } catch (Exception e) {
                logger.error("Unable to extract a configuration for CMI", e);
                throw new ConfigurationException(
                        "Unable to extract a configuration for CMI", e);
            }
        }
        return configure(sc, domainName, serverName, agentId, cmiConfig);
    }

    /**
     * @param sc
     * @param domainName
     * @param serverName
     * @param agentId
     * @param cmiEnv
     * @return
     * @throws ConfigurationException
     */
    protected static boolean enableCMI(final ServerConfigurationImpl sc,
            final String domainName, final String serverName, final String agentId,
            final Properties env) throws ConfigurationException {
        CMIConfig<?> cmiConfig;
        Properties cmiEnv;
        if(env == null || env.isEmpty()) {
            ProtocolConfiguration[] protocolConfigurations = ConfigurationRepository.getConfigurations();
            if(protocolConfigurations.length != 1) {
                logger.debug("Many protocols: CMI client only supports one protocol so you need to specify one of them.");
                return false;
            }
            cmiEnv = getDefaultCMIEnv(protocolConfigurations[0].getProtocol().getName());
        } else {
            cmiEnv = env;
        }
        try {
            cmiConfig = CMIConfigurator.extractCMIConfig(cmiEnv, false);
        } catch (CMIConfigException e) {
            logger.error("Unable to extract a configuration for CMI", e);
            throw new ConfigurationException(
                    "Unable to extract a configuration for CMI", e);
        }
        return configure(sc, domainName, serverName, agentId, cmiConfig);
    }

    private static Properties getDefaultCMIEnv(final String protocolName) {
        Properties cmiEnv = new Properties();
        cmiEnv.setProperty(CMIProperty.WRAPPED_PROTOCOL.getPropertyName(), protocolName);
        return cmiEnv;
    }

    @SuppressWarnings("unchecked")
    private static boolean configure(final ServerConfiguration sc, final String domainName,
            final String serverName, final String agentId, final CMIConfig<?> cmiConfig) throws ConfigurationException {
        ClusterViewManagerFactoryConfig<?> clusterViewManagerFactoryConfig =
            cmiConfig.getClusterViewManagerFactoryConfig();
        if(clusterViewManagerFactoryConfig == null) {
            org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig clientClusterViewManagerFactoryConfig =
                new org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig();
            clusterViewManagerFactoryConfig = clientClusterViewManagerFactoryConfig;
            ((CMIConfig<org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig>) cmiConfig)
                .setClusterViewManagerFactoryConfig(clientClusterViewManagerFactoryConfig);
        }
        boolean cmiClient =
            clusterViewManagerFactoryConfig instanceof org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig;
        if (cmiClient) {
            if(cmiConfig.isNoCmi()) {
                logger.info("CMI disabled by the XML configuration file.");
                return false;
            }
            if(clusterViewManagerFactoryConfig.getConfig() == null) {
                ((org.ow2.cmi.controller.factory.client.ClusterViewManagerFactoryConfig) clusterViewManagerFactoryConfig).setConfig(
                        new org.ow2.cmi.controller.client.Config());
            }
        }
        try {
            if(!enableCMIOnEachProtocol(sc, cmiConfig, cmiClient)) {
                return false;
            }
        } catch(Exception e) {
            if(sc.isServer()) {
                logger.error("CMI cannot be enabled", e);
                throw new ConfigurationException("CMI cannot be enabled", e);
            }
            logger.debug("CMI cannot be enabled", e);
            return false;
        }
        if (sc.isServer() && cmiConfig.getClusterViewManagerFactoryConfig().getConfig().isAdminEnabled()) {
            MBeanUtils.setMBeanParameters(domainName, serverName, agentId);
        }
        try {
            CMIConfigurator.configure(
                    ClusterViewManagerFactory.getFactory(), cmiConfig);
        } catch (Exception e) {
            logger.error("Unable to configure CMI", e);
            throw new ConfigurationException(
                    "Unable to configure CMI", e);
        }
        return true;
    }

    /**
     * Enable CMI on the protocols according to the given CMI configuration.
     * @param cmiConfig CMI configuration to check the activation of each protocol
     * @throws ConfigurationException
     */
    private static boolean enableCMIOnEachProtocol(
            final ServerConfiguration sc,
            final CMIConfig<?> cmiConfig,
            final boolean cmiClient) throws ConfigurationException {
        boolean cmiEnabled = false;

        for (ProtocolConfiguration protocol : ConfigurationRepository.getConfigurations()) {
            cmiEnabled |= enableCmiOnProtocol((ProtocolConfigurationImpl) protocol, sc, cmiConfig, cmiClient);
        }
        return cmiEnabled;
    }

    /**
     * Enable CMI for this protocol.
     * @throws ConfigurationException
     */
    private static boolean enableCmiOnProtocol(
            final ProtocolConfigurationImpl protocolConfiguration,
            final ServerConfiguration sc,
            final CMIConfig<?> cmiConfig,
            final boolean cmiClient) throws ConfigurationException {

        String protocolName = protocolConfiguration.getProtocol().getName();
        Properties cmiEnv = CMIConfigurationHelper.extractCMIProperties(
                protocolConfiguration, sc.isServer(), cmiConfig, cmiClient);
        List<String> providerUrls = null;
        String clusterViewProviderUrls = cmiEnv.getProperty(CMIProperty.CLUSTER_VIEW_PROVIDER_URLS.getPropertyName());
        if (cmiClient && clusterViewProviderUrls == null) {
            logger.error("Provider URL missing");
            throw new ConfigurationException("Provider URL missing");
        }

        if(clusterViewProviderUrls != null) {
            try {
                providerUrls =
                    ProviderURLsParser.parse(protocolName, clusterViewProviderUrls);
            } catch (Exception e) {
                logger.debug("Unable to activate CMI", e);
                throw new ConfigurationException("Unable to activate CMI", e);
            }
        }
        if(cmiClient && !sc.isServer()) {
            if(!checkCmiEnabling(
                    protocolConfiguration.getProtocol(),
                    providerUrls,
                    ((IConfig) cmiConfig.getClusterViewManagerFactoryConfig().getConfig()).getClientProviderName())) {
                return false;
            }
        }
        JNDIConfig jndiConfig = cmiConfig.getJNDIConfig();
        ProtocolConfig protocolConfig = null;
        if (!cmiClient) {
            // only check that the protocol is enabled
            if (jndiConfig != null) {
                Set<ProtocolConfig> protocolConfigs = jndiConfig.getProtocolConfigs();
                for(ProtocolConfig protoConfig : protocolConfigs) {
                    if(protocolName.equals(protoConfig.getName())) {
                        protocolConfig = protoConfig;
                        break;
                    }
                }
                if (protocolConfig != null && protocolConfig.isNoCmi()) {
                    return false;
                }
            }
        } else {
            // Fill the protocol field in order to configure the CMI client
            if (jndiConfig == null) {
                jndiConfig = new JNDIConfig();
                cmiConfig.setJNDIConfig(jndiConfig);
            } else {
                Set<ProtocolConfig> protocolConfigs = jndiConfig.getProtocolConfigs();
                for(ProtocolConfig protoConfig : protocolConfigs) {
                    if(protocolName.equals(protoConfig.getName())) {
                        protocolConfig = protoConfig;
                        break;
                    }
                }
            }
            if (protocolConfig == null) {
                protocolConfig = new ProtocolConfig();
                jndiConfig.addProtocolConfig(protocolConfig);
            }
            protocolConfig.setName(protocolName);
            protocolConfig.setInitialContextFactoryName(
                    protocolConfiguration.getProtocol().getInitialContextFactoryClassName());
            if (providerUrls != null) {
                ProviderUrls pUrls = new ProviderUrls();
                pUrls.setProviderUrls(providerUrls);
                protocolConfig.setProviderUrls(pUrls);
            }
            protocolConfig.setLocalRegistryUrl(protocolConfiguration.getProviderURL());
        }
        protocolConfiguration.enableCmi(cmiEnv);
        return true;
    }

    /**
     * Extract CMI properties of properties.
     * @throws ConfigurationException if properties are missing
     */
    private static Properties extractCMIProperties(
            final ProtocolConfigurationImpl protocolConfiguration,
            final boolean isServer,
            final CMIConfig<?> cmiConfig,
            final boolean cmiClient) throws ConfigurationException {

        String protocolName = protocolConfiguration.getProtocol().getName();
        String prefixProtocol = CarolDefaultValues.CAROL_PREFIX + "." + protocolName + ".";

         Properties cmiEnv = new Properties();

        // Firstly fix the properties specific at CMI
        cmiEnv.setProperty(CMIProperty.WRAPPED_PROTOCOL.getPropertyName(), protocolName);

        String carolProviderUrl = protocolConfiguration.getValue(prefixProtocol + CarolDefaultValues.URL_PREFIX);

        String clusterViewProviderURLs = null;

        if(isServer) {
            cmiEnv.setProperty(
                    CMIProperty.LOCAL_REGISTRY_URL.getPropertyName(), carolProviderUrl);
        }
        JNDIConfig jndiConfig = cmiConfig.getJNDIConfig();
        if(jndiConfig != null) {
            Set<ProtocolConfig> protocolConfigs = jndiConfig.getProtocolConfigs();
            ProtocolConfig currentProtocolConfig = null;
            String protocol = JNDIConfig.getCurrentProtocolName();
            if (protocol != null) {
                for(ProtocolConfig protocolConfig : protocolConfigs) {
                    if(protocolConfig.getName().equals(protocol)) {
                        currentProtocolConfig = protocolConfig;
                        break;
                    }
                }
            } else if (protocolConfigs.size() == 1
                    && ConfigurationRepository.getConfigurations().length == 1) {
                currentProtocolConfig = (ProtocolConfig) protocolConfigs.toArray()[0];
                if(currentProtocolConfig.getName() == null ||
                        currentProtocolConfig.getName().equals(protocolName)) {
                    currentProtocolConfig.setName(protocolName);
                } else {
                    currentProtocolConfig = null;
                }
            }
            if (currentProtocolConfig != null) {
                ProviderUrls pUrls = currentProtocolConfig.getProviderUrls();
                List<String> urls = new ArrayList<String>();
                if (pUrls != null) {
                    urls.addAll(pUrls.getProviderUrls());
                }
                if (!isServer) {
                    urls.add(carolProviderUrl);
                }
                if (!urls.isEmpty()) {
                    clusterViewProviderURLs = ProviderURLsParser.parse(urls);
                }
            }
        }
        if (clusterViewProviderURLs != null) {
            cmiEnv.setProperty(CMIProperty.CLUSTER_VIEW_PROVIDER_URLS.getPropertyName(), clusterViewProviderURLs);
        }
        cmiEnv.setProperty(CMIProperty.WRAPPED_INITIAL_CONTEXT_FACTORY.getPropertyName(),
                protocolConfiguration.getProtocol().getInitialContextFactoryClassName());
        cmiEnv.setProperty(CMIProperty.SERVER_MODE.getPropertyName(), Boolean.toString(!cmiClient));

        // Finally fix the JNDI properties to use CMI
        cmiEnv.setProperty(Context.INITIAL_CONTEXT_FACTORY, CMIInitialContextFactory.class.getName());

        return cmiEnv;
    }

    private static boolean checkCmiEnabling(
            final Protocol protocol, final List<String> providerUrls, final String providerName) {
        Hashtable<String, Object> env = new Hashtable<String, Object>();
        env.put(Context.INITIAL_CONTEXT_FACTORY, protocol.getInitialContextFactoryClassName());
        for(String providerUrl : providerUrls) {
            env.put(Context.PROVIDER_URL, providerUrl);
            Context context = null;
            try {
                context = new InitialContext(env);
                context.lookup(providerName);
                return true;
            } catch (NamingException e) {
                logger.debug("Unable to use CMI with the provider URL " + providerUrl, e);
            } finally {
                if(context != null) {
                    try {
                        context.close();
                    } catch (NamingException e) {
                        logger.debug("Unable to close the context", e);
                    }
                }
            }
        }
        return false;
    }

}
