/**
 * Copyright (C) 2005-2009 - Bull S.A.S.
 * Contact: carol@ow2.org
 *
 * CAROL: Common Architecture for RMI ObjectWeb Layer
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 *
 * --------------------------------------------------------------------------
 * $Id: ConfigurationUtil.java 2206 2010-02-24 16:21:55Z benoitf $
 * --------------------------------------------------------------------------
 */
package org.ow2.carol.util.configuration;


/**
 * Utility class to analyze some configuration properties
 * @author Florent Benoit
 */
public class ConfigurationUtil {

    /**
     * Utility class, no public constructor
     */
    private ConfigurationUtil() {

    }


    /**
     * Parses the given url, and returns the port number. 0 is given in error
     * case)
     * @param url given url on which extract port number
     * @return port number of the url
     * @throws ConfigurationException if URL is invalid
     */
    public static int getPortOfUrl(final String url) throws ConfigurationException {
        int portNumber = 0;
        int closingBrac = -1;

        // Keep only <host>:<port>
        String hostPort = extractHostPort(url);

        if (hostPort.startsWith("[")) {	// at IPv6 literal
            closingBrac = hostPort.indexOf(']', 1);
            if (closingBrac < 0) {
                throw new ConfigurationException(
                        "Missing the closing bracket in the following IPv6 address: " + hostPort);
            }
        }
        int m = hostPort.indexOf(':', closingBrac + 1);

        try {
            if (m > -1) {
                if (m == 0) { // no host
                    if (hostPort.length() > 1) {
                        // non-empty port
                        portNumber = Integer.parseInt(hostPort.substring(1));
                    }
                } else { // non-empty host, maybe non-empty port
                    if (hostPort.length() > m + 1) { // port specified
                        portNumber = Integer.parseInt(hostPort.substring(m + 1));
                    }
                }
            }
        } catch (Exception e) {
            // don't rethrow original exception. only URL name is important
            throw new ConfigurationException(
                    "Invalid URL '" + url + "'. It should be on the format <protocol>://<hostname>:<port>");
        }
        return portNumber;
    }

    /**
     * Parses the given url, and returns the hostname
     * If not found, returns localhost
     * @param url given url on which extract hostname
     * @return hostname of the url
     * @throws ConfigurationException if URL is invalid
     */
    public static String getHostOfUrl(final String url) throws ConfigurationException {
        String host = null;
        int closingBrac = -1;

        // Keep only <host>:<port>
        String hostPort = extractHostPort(url);

        if (hostPort.startsWith("[")) {	// at IPv6 literal
            closingBrac = hostPort.indexOf(']', 1);
            if (closingBrac < 0) {
                throw new ConfigurationException(
                        "Missing the closing bracket in the following IPv6 address: " + hostPort);
            }
        }
        int m = hostPort.indexOf(':', closingBrac + 1);

        try {
            if (m > 0) {
                host = hostPort.substring(0, m);
            }
        } catch (Exception e) {
            // don't rethrow original exception. only URL name is important
            throw new ConfigurationException(
                    "Invalid URL '" + url + "'. It should be on the format <protocol>://<hostname>:<port>");
        }
        return host;
    }

    private static String extractHostPort(final String url) throws ConfigurationException {
        String hostPort;

        // Skip "<protocol>://"
        hostPort = url.substring(url.indexOf("://") + 3);

        // Skip "/*"
        int slash = hostPort.indexOf("/");

        if (slash > -1) {
            hostPort = hostPort.substring(0, slash);
        }

        // Skip ",*"
        int comma = hostPort.indexOf(",");

        if (comma > -1) {
            hostPort = hostPort.substring(0, comma);
        }
        return hostPort;
    }
}
